"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gi.repository import Gtk, GObject


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/account_edit.ui')
class EditAccountWindow(Gtk.Window):
    __gtype_name__ = 'EditAccountWindow'
    # Signals
    __gsignals__ = {
        'updated': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str, GObject.TYPE_PYOBJECT, )
        ),
    }
    # Widgets
    save_btn: Gtk.Button = Gtk.Template.Child()
    back_btn: Gtk.Button = Gtk.Template.Child()

    def __init__(self, account):
        super(EditAccountWindow, self).__init__()
        self.init_template('EditAccountWindow')

        self._account = account
        self.__init_widgets()

    def __init_widgets(self):
        from .add import AccountConfig

        self.account_config = AccountConfig(edit=True, account=self._account)
        self.account_config.connect("changed", self._on_account_config_changed)

        self.back_btn.connect("clicked", lambda *_: self.destroy())

        self.add(self.account_config)

    def _on_account_config_changed(self, _, state: bool):
        """
        Set the sensitivity of the AddButton
            depends on the AccountConfig.

        :param state: the state of the save button
        :type state: bool
        """
        self.save_btn.set_sensitive(state)

    @Gtk.Template.Callback('save_btn_clicked')
    def _on_save(self, *_):
        """
            Save Button clicked signal handler.
        """
        from .list import AccountsWidget
        ac_widget = AccountsWidget.get_default()

        new_account = self.account_config.account
        username = new_account["username"]
        provider = new_account["provider"]
        old_provider = self._account.provider
        # Update the AccountRow widget
        self.emit("updated", username, provider)
        # Update the providers list
        if provider.provider_id != old_provider.provider_id:
            ac_widget.update_provider(self._account, provider)

        ac_widget.update_provider_image(provider)
        self.destroy()



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
import logging


class Logger:
    """Logging handler."""
    DEBUG: int = logging.DEBUG
    ERROR: int = logging.ERROR
    # Default instance of Logger
    instance: 'Logger' = None
    # Message format
    FORMAT: str = "[%(levelname)-s] %(asctime)s %(message)s"
    # Date format
    DATE: str = "%Y-%m-%d %H:%M:%S"

    def __init__(self):
        pass

    @staticmethod
    def new():
        """Create a new instance of Logger."""
        logger = logging.getLogger('authenticator')
        handler = logging.StreamHandler()
        formatter = logging.Formatter(Logger.FORMAT, Logger.DATE)
        handler.setFormatter(formatter)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        return logger

    @staticmethod
    def get_default() -> 'Logger':
        """Return the default instance of Logger."""
        if Logger.instance is None:
            # Init the Logger
            Logger.instance = Logger.new()
        return Logger.instance

    @staticmethod
    def set_level(level: int):
        """Set the logging level."""
        Logger.get_default().setLevel(level)

    @staticmethod
    def warning(msg: str):
        """Log a warning message."""
        Logger.get_default().warning(msg)

    @staticmethod
    def debug(msg: str):
        """Log a debug message."""
        Logger.get_default().debug(msg)

    @staticmethod
    def info(msg: str):
        """Log an info message."""
        Logger.get_default().info(msg)

    @staticmethod
    def error(msg: str):
        """Log an error message."""
        Logger.get_default().error(msg)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gi.repository import GObject, GLib
from typing import Iterable

from .account import Account
from .database import Database
from .provider import Provider


class AccountsManager(GObject.GObject):
    __gsignals__ = {
        'counter_updated': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (int,)
        ),
    }

    instance: 'AccountsManager' = None
    empty: GObject.Property = GObject.Property(type=bool, default=False)

    # A list that contains a tuple (provider, accounts)
    __accounts = []

    def __init__(self):
        GObject.GObject.__init__(self)
        self._accounts_per_provider = []
        self._alive = True
        self.__fill_accounts()
        self._timeout_id = 0
        self.counter_max = 30
        self.counter = self.counter_max
        self._start_progress_countdown()

    @staticmethod
    def get_default() -> 'AccountsManager':
        if AccountsManager.instance is None:
            AccountsManager.instance = AccountsManager()
        return AccountsManager.instance

    def add(self, provider: 'Provider', account: 'Account'):
        added = False
        for _provider, accounts in self._accounts_per_provider:
            if provider == _provider:
                accounts.append(account)
                added = True
                break
        if not added:
            self._accounts_per_provider.append((provider, [account]))
        self.props.empty = False
        if self._timeout_id == 0:
            self._start_progress_countdown()

    def delete(self, account: 'Account'):
        provider_index = 0
        _accounts = None
        for provider, accounts in self._accounts_per_provider:
            if account in accounts:
                _accounts = accounts
                break
            provider_index += 1
        if provider:
            _accounts.remove(account)
            if not len(_accounts):
                del self._accounts_per_provider[provider_index]
        self.props.empty = len(self._accounts_per_provider) == 0
        if self.props.empty:
            self._stop_progress_countdown()

    def search(self, terms: Iterable[str]):
        accounts = Database.get_default().search_accounts(terms)
        _accounts = []
        for account in accounts:
            account = Account(*account)
            if account.otp:
                _accounts.append(account)
        return _accounts

    @property
    def accounts_per_provider(self):
        return self._accounts_per_provider

    @property
    def accounts(self):
        accounts = []
        for _, _accounts in self._accounts_per_provider:
            accounts.extend(_accounts)
        return accounts

    @property
    def accounts_count(self):
        count = 0
        for _, accounts in self._accounts_per_provider:
            count += len(accounts)
        return count

    def kill(self):
        self._alive = False
        self._stop_progress_countdown()

    def update_childes(self, signal: str):
        for _, accounts in self._accounts_per_provider:
            for account in accounts:
                account.emit(signal)

    def __update_counter(self, *args):
        if self._alive:
            self.counter -= 1
            if self.counter == 0:
                self.counter = self.counter_max
                self.update_childes("otp_out_of_date")
            self.emit("counter_updated", self.counter)
            return True
        return False

    def __fill_accounts(self):
        providers = Database.get_default().get_providers(only_used=True)
        for provider in providers:
            accounts = Database.get_default().accounts_by_provider(provider.id)
            provider = Provider(*provider)
            _accounts = []
            for account in accounts:
                account = Account(*account)
                if account.otp:
                    _accounts.append(account)
            self._accounts_per_provider.append((provider, _accounts))
        self.props.empty = len(self._accounts_per_provider) == 0

    def _start_progress_countdown(self):
        self._stop_progress_countdown()
        self._timeout_id = GLib.timeout_add_seconds(1, self.__update_counter,
                                                    None)

    def _stop_progress_countdown(self):
        if self._timeout_id > 0:
            GLib.Source.remove(self._timeout_id)
            self._timeout_id = 0
        self.counter = self.counter_max



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from os import path
from tempfile import NamedTemporaryFile

from gi.repository import Gio, GLib


class GNOMEScreenshot:
    """
        GNOME Screenshot interface implementation.
        Currently implements the only needed method by Authenticator.
    """
    interface = "org.gnome.Shell.Screenshot"
    path = "/org/gnome/Shell/Screenshot"

    def __init__(self):
        pass

    @staticmethod
    def area(filename: str = None) -> str:
        """
            Take a screen shot of an area and save it to a specific filename
            Using GNOME Shell Screenshot Interface.
            :param filename: output filename
            :type filename: str
        """
        if not filename:
            filename = path.join(GLib.get_user_cache_dir(),
                                 path.basename(NamedTemporaryFile().name))
        bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        screen_proxy = Gio.DBusProxy.new_sync(bus,
                                              Gio.DBusProxyFlags.NONE,
                                              None,
                                              GNOMEScreenshot.interface,
                                              GNOMEScreenshot.path,
                                              GNOMEScreenshot.interface,
                                              None)
        x, y, width, height = screen_proxy.call_sync('SelectArea', None, Gio.DBusCallFlags.NONE,
                                                     -1, None).unpack()

        args = GLib.Variant('(iiiibs)', (x, y, width, height, False, filename
                                         )
                            )
        screenshot = screen_proxy.call_sync('ScreenshotArea', args,
                                            Gio.DBusCallFlags.NONE, -1, None)
        success, filename = screenshot.unpack()
        return filename if success else None



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from .keyring import Keyring
from .logger import Logger
from .otp import OTP

from .qr_reader import QRReader
from .screenshot import GNOMEScreenshot
from .settings import Settings

from .database import Database
from .provider import Provider
from .account import Account
from .accounts_manager import AccountsManager
from .backup import BackupJSON



"""
Empty uneeded provider iamges
"""
from os import path
from yoyo import step

__depends__ = {'authenticator_20190525_04_Faezz-restore-old-accounts'}


def do_step(conn):

    providers_db = conn.execute("SELECT id, image FROM providers").fetchall()

    to_empty = []
    for provider_id, provider_image in providers_db:
        if provider_image and path.basename(provider_image) == provider_image:
            to_empty.append(str(provider_id))

    if len(to_empty):
        providers_ids = ", ".join(to_empty)
        conn.execute("UPDATE providers SET image='' WHERE id IN (?)", (providers_ids, ))


steps = [
    step(do_step, ignore_errors='apply')
]



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gettext import gettext as _
from gi.repository import Gio, Gtk, GObject, Handy
from Authenticator.models import Settings, Keyring

__all__ = ['SettingsWindow']


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/settings.ui')
class SettingsWindow(Handy.PreferencesWindow):
    __gtype_name__ = 'SettingsWindow'

    dark_theme_switch: Gtk.Switch = Gtk.Template.Child()
    night_light_switch: Gtk.Switch = Gtk.Template.Child()

    lock_row: Handy.ExpanderRow = Gtk.Template.Child()

    lock_timeout_row: Handy.ActionRow = Gtk.Template.Child()
    lock_timeout_spinbtn: Gtk.SpinButton = Gtk.Template.Child()

    def __init__(self):
        super(SettingsWindow, self).__init__()
        self.init_template('SettingsWindow')

        self.__init_widgets()
        self.__bind_signals()

    def __init_widgets(self):

        self._password_widget = PasswordWidget()
        self._password_widget.parent = self
        self.lock_row.add(self._password_widget)

    def _on_lock_row_expanded(self, *_):
        keyring = Keyring.get_default()
        if keyring.has_password():
            keyring.set_password_state(self.lock_row.props.expanded)
            self.lock_row_toggle_btn.props.active = False

    def __on_lock_switch_toggled(self, toggle_btn: Gtk.ToggleButton, *_):
        toggled = toggle_btn.props.active
        expansion_enabled = self.lock_row.props.enable_expansion
        if not Keyring.get_default().has_password() and not toggled and expansion_enabled:
            self.lock_row.props.enable_expansion = False

    def __bind_signals(self):
        settings = Settings.get_default()
        self.dark_theme_switch.set_active(settings.dark_theme and not settings.night_light)

        self.night_light_switch.set_active(settings.night_light)
        settings.bind("night-light", self.night_light_switch,
                      "active", Gio.SettingsBindFlags.DEFAULT)

        keyring = Keyring.get_default()
        # Hackish solution to get the expander from HdyExpanderRow
        self.lock_row.props.enable_expansion = keyring.has_password()
        self.lock_row_toggle_btn = self.lock_row.get_children()[0].get_children()[3]

        self.lock_row.props.enable_expansion = Keyring.get_default().is_password_enabled()
        self.lock_row.connect("notify::enable-expansion", self.__on_enable_password)
        self.lock_row_toggle_btn.connect("notify::active", self.__on_lock_switch_toggled)
        self.lock_row.connect("notify::expanded", self._on_lock_row_expanded)

        keyring.bind_property("can-be-locked", self.lock_timeout_row, "sensitive",
                              GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE)
        self.lock_timeout_spinbtn.props.value = settings.auto_lock_timeout
        settings.bind("auto-lock-timeout", self.lock_timeout_spinbtn, "value",
                      Gio.SettingsBindFlags.DEFAULT)

        self._password_widget.connect("password-updated", self.__on_password_updated)
        self._password_widget.connect("password-deleted", self.__on_password_deleted)

        def on_night_light_switch(switch: Gtk.Switch, _):
            # Set the application to use Light theme
            if switch.get_active() and self.dark_theme_switch.get_active():

                self.dark_theme_switch.set_active(False)

        self.night_light_switch.connect("notify::active", on_night_light_switch)

        def on_dark_theme_switch(switch: Gtk.Switch, _):
            # Set the application to use Light theme

            if settings.night_light and switch.get_active():
                switch.set_state(False)
            elif not settings.night_light:
                settings.dark_theme = switch.get_active()

        self.dark_theme_switch.connect("notify::active", on_dark_theme_switch)

    def __on_enable_password(self, *_):
        keyring = Keyring.get_default()
        keyring.set_password_state(self.lock_row.props.enable_expansion)
        if not keyring.has_password():
            self._password_widget.set_current_password_visibility(False)
        else:
            self._password_widget.set_current_password_visibility(True)

    def __on_password_updated(self, *_):
        self.lock_row_toggle_btn.props.active = False

    def __on_password_deleted(self, *__):
        # self.notification.send(_("The authentication password was deleted."))
        self.lock_row.set_enable_expansion(False)
        self.lock_row_toggle_btn.props.active = False


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/password_widget.ui')
class PasswordWidget(Gtk.Box):
    __gtype_name__ = 'PasswordWidget'
    __gsignals__ = {
        'password-updated': (
            GObject.SignalFlags.RUN_LAST,
            None,
            ()
        ),
        'password-deleted': (
            GObject.SignalFlags.RUN_LAST,
            None,
            ()
        ),
    }

    delete_password_btn: Gtk.Button = Gtk.Template.Child()
    change_password_btn: Gtk.Button = Gtk.Template.Child()

    password_entry: Gtk.Entry = Gtk.Template.Child()
    confirm_password_entry: Gtk.Entry = Gtk.Template.Child()
    current_password_entry: Gtk.Entry = Gtk.Template.Child()

    current_password_box: Gtk.Box = Gtk.Template.Child()

    def __init__(self):
        super(PasswordWidget, self).__init__()
        self.parent = None
        self.init_template('PasswordWidget')

    def reset_widgets(self):
        """Reset widgets state."""
        self.password_entry.set_text("")
        self.confirm_password_entry.set_text("")
        self.current_password_entry.set_text("")

        self.password_entry.get_style_context().remove_class("error")
        self.confirm_password_entry.get_style_context().remove_class("error")
        self.current_password_entry.get_style_context().remove_class("error")
        self.change_password_btn.set_sensitive(False)

    def set_current_password_visibility(self, visibilty: bool):
        if not visibilty:
            self.current_password_box.hide()
            self.delete_password_btn.hide()
            self.change_password_btn.set_label(_("Save Password"))
        else:
            self.current_password_box.show()
            self.delete_password_btn.show()
            self.change_password_btn.set_label(_("Change Password"))

    @Gtk.Template.Callback('password_entry_changed')
    def __validate_password(self, *_):
        keyring = Keyring.get_default()
        password = self.password_entry.get_text()
        repeat_password = self.confirm_password_entry.get_text()
        if not password:
            self.password_entry.get_style_context().add_class("error")
            valid_password = False
        else:
            self.password_entry.get_style_context().remove_class("error")
            valid_password = True

        if not repeat_password or password != repeat_password:
            self.confirm_password_entry.get_style_context().add_class("error")
            valid_repeat_password = False
        else:
            self.confirm_password_entry.get_style_context().remove_class("error")
            valid_repeat_password = True
        to_validate = [valid_password, valid_repeat_password]

        if keyring.has_password():
            old_password = self.current_password_entry.get_text()
            if old_password != keyring.get_password():
                self.current_password_entry.get_style_context().add_class("error")
                valid_old_password = False
            else:
                self.current_password_entry.get_style_context().remove_class("error")
                valid_old_password = True
            to_validate.append(valid_old_password)

        self.change_password_btn.set_sensitive(all(to_validate))

    @Gtk.Template.Callback('update_password_clicked')
    def __save_password(self, *__):
        if self.change_password_btn.get_sensitive():
            keyring = Keyring.get_default()
            password = self.password_entry.get_text()
            keyring.set_password(password)
            self.reset_widgets()
            self.set_current_password_visibility(True)
            self.emit("password-updated")

    @Gtk.Template.Callback('reset_password_clicked')
    def __reset_password(self, *args):
        dialog = Gtk.MessageDialog(buttons=Gtk.ButtonsType.YES_NO)
        dialog.props.message_type = Gtk.MessageType.QUESTION
        dialog.props.text = _("Do you want to remove the authentication password?")
        dialog.props.secondary_text = _("Authentication password enforces the privacy of your accounts.")

        dialog.set_transient_for(self.parent)

        response = dialog.run()
        if response == Gtk.ResponseType.YES:
            Keyring.get_default().remove_password()
            self.reset_widgets()
            self.set_current_password_visibility(False)
            self.emit("password-deleted")
        dialog.destroy()



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gettext import gettext as _
from gi.repository import Gtk, GObject

from Authenticator.widgets.provider_image import ProviderImage
from .row import AccountRow
from Authenticator.models import Account, AccountsManager, Provider


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/accounts_widget.ui')
class AccountsWidget(Gtk.Box):
    __gtype_name__ = 'AccountsWidget'

    __gsignals__ = {
        'account-removed': (GObject.SignalFlags.RUN_LAST, None, ()),
        'account-added': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    instance: 'AccountsWidget' = None

    accounts_container = Gtk.Template.Child()
    otp_progress_bar = Gtk.Template.Child()

    def __init__(self):
        super(AccountsWidget, self).__init__()
        self.init_template('AccountsWidget')

        self._providers = []
        self._to_delete = []
        self.__init_widgets()

    def __init_widgets(self):
        accounts_manager = AccountsManager.get_default()
        accounts_manager.connect("counter_updated",
                                 self._on_counter_updated)
        # Add different accounts to the main view
        for provider, accounts in accounts_manager.accounts_per_provider:
            for account in accounts:
                self.append(account)

    def __add_provider(self, provider):
        accounts_list = self._get_by_provider(provider)["accounts_list"]
        if not accounts_list:
            accounts_list = AccountsList()
            accounts_list.connect("account-deleted", self._on_account_deleted)
            self._providers.append({"provider": provider,
                                    "accounts_list": accounts_list})
            provider_widget = ProviderWidget(accounts_list, provider)
            self.accounts_container.pack_start(provider_widget, False, False, 0)
        return accounts_list

    def _get_by_provider(self, provider):
        for provider_info in self._providers:
            if provider.provider_id == provider_info['provider'].provider_id:
                return provider_info
        return {
            'provider': None,
            'accounts_list': None
        }

    @staticmethod
    def get_default() -> 'AccountsWidget':
        """Return the default instance of AccountsWidget."""
        if AccountsWidget.instance is None:
            AccountsWidget.instance = AccountsWidget()
        return AccountsWidget.instance

    def append(self, account):
        accounts_list = self.__add_provider(account.provider)
        accounts_list.add_row(account)

        self._reorder()
        self.emit("account-added")

    @property
    def accounts_lists(self):
        return [provider['accounts_list'] for provider in self._providers]

    def update_provider_image(self, provider):
        for child in self.accounts_container.get_children():
            if child.provider.provider_id == provider.provider_id:
                child.provider_image.set_image(provider.image)
                break

    def update_provider(self, account, new_provider):
        current_account_list = None
        account_row = None
        for account_list in self.accounts_lists:
            for account_row in account_list:
                if account_row.account == account:
                    current_account_list = account_list
                    break

            if current_account_list:
                break
        if account_row:
            current_account_list.remove(account_row)
            account_row.account.provider = new_provider
            self.append(account_row.account)
        self._on_account_deleted(current_account_list, None)
        self._reorder()
        self._clean_unneeded_providers_widgets()

    def _on_account_deleted(self, accounts_list, account=None):
        if account:
            AccountsManager.get_default().delete(account)
        if len(accounts_list.get_children()) == 0:
            self._to_delete.append(accounts_list)
        self._reorder()
        self._clean_unneeded_providers_widgets()
        self.emit("account-removed")

    def _clean_unneeded_providers_widgets(self):
        for accounts_list in self._to_delete:
            provider_widget = accounts_list.get_parent()
            self.accounts_container.remove(provider_widget)
            self._providers.remove(self._get_by_provider(provider_widget.provider))
        self._to_delete = []

    def _reorder(self):
        """
            Re-order the ProviderWidget on AccountsWidget.
        """
        childs = self.accounts_container.get_children()
        ordered_childs = sorted(
            childs, key=lambda children: children.provider.name.lower())
        for i in range(len(ordered_childs)):
            self.accounts_container.reorder_child(ordered_childs[i], i)
        self.show_all()

    def _on_counter_updated(self, accounts_manager, counter):
        counter_fraction = counter / accounts_manager.counter_max
        self.otp_progress_bar.set_fraction(counter_fraction)
        self.otp_progress_bar.set_tooltip_text(
            _("The One-Time Passwords expire in {} seconds").format(counter))


class ProviderWidget(Gtk.Box):

    def __init__(self, accounts_list: Gtk.ListBox, provider: Provider):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.get_style_context().add_class("provider-widget")
        self.provider = provider
        self.accounts_list = accounts_list
        self._build_widgets()

    def _build_widgets(self):
        self.set_valign(Gtk.Align.START)

        provider_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        provider_lbl = Gtk.Label()
        provider_lbl.set_text(self.provider.name)
        provider_lbl.set_halign(Gtk.Align.START)
        provider_lbl.get_style_context().add_class("provider-name")

        self.provider_image = ProviderImage(self.provider, 48)

        provider_container.pack_start(self.provider_image, False, False, 6)
        provider_container.pack_start(provider_lbl, False, False, 6)

        self.pack_start(provider_container, False, False, 6)
        self.pack_start(self.accounts_list, False, False, 6)


class AccountsList(Gtk.ListBox):
    """Accounts List."""

    __gsignals__ = {
        'account-deleted': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (Account, )
        ),
        'account-added': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (Account, )
        )
    }

    def __init__(self):
        Gtk.ListBox.__init__(self)
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.get_style_context().add_class("accounts-list")
        self.get_style_context().add_class("frame")
        self.set_header_func(self._update_header_func)

    def add_row(self, account: Account):
        row = AccountRow(account)
        row.delete_btn.connect("clicked", self.__on_delete_child, row)
        self.add(row)
        self.emit("account-added", account)

    def __on_delete_child(self, _, account_row):
        self.remove(account_row)
        account = account_row.account
        account.remove()
        self.emit("account-deleted", account)

    def _update_header_func(self, row, row_before):
        def on_realize_sep(separator):
            separator.set_size_request(row_before.get_allocated_width(), -1)

        if row_before:
            separator = Gtk.Separator()
            separator.connect("realize", on_realize_sep)
            row.set_header(separator)
            separator.show()



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""

from .notification import Notification
from .provider_image import ProviderImage
from .utils import export_json, import_json
from .window import Window, WindowView

from .accounts.add import AddAccountWindow
from .accounts.edit import EditAccountWindow
from .accounts.list import AccountsWidget

from .settings import SettingsWindow


"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from os import remove, path
from urllib.parse import urlparse, parse_qsl, unquote

from PIL import Image
from pyzbar.pyzbar import decode

from Authenticator.models import Logger, OTP


class QRReader:

    @staticmethod
    def from_file(filename: str):
        decoded_data = decode(Image.open(filename))
        if path.isfile(filename):
            remove(filename)
        try:
            # See https://github.com/google/google-authenticator/wiki/Key-Uri-Format
            # for a description of the URL format
            url = urlparse(decoded_data[0].data.decode())
            query_params = parse_qsl(url.query)
            url_data = dict(query_params)

            username = None
            label = unquote(url.path.lstrip("/"))
            if ":" in label:
                provider, username = label.split(":", maxsplit=1)
            else:
                provider = label
            # provider information could also be in the query params
            provider = url_data.get("issuer", provider)

            token = url_data.get("secret")
            assert OTP.is_valid(token)

            return {
                'username': username,
                'provider': provider,
                'token': token
            }
        except (KeyError, IndexError):
            Logger.error("Invalid QR image")



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gettext import gettext as _
from gi.repository import Gdk, Gtk, GObject, Gio, Handy

from .list import AccountsWidget
from Authenticator.widgets.notification import Notification
from Authenticator.widgets.provider_image import ProviderImage, ProviderImageState
from Authenticator.models import AccountsManager, Account, OTP, Provider, QRReader, GNOMEScreenshot


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/account_add.ui')
class AddAccountWindow(Gtk.Window):
    """Add Account Window."""
    __gtype_name__ = "AddAccountWindow"
    # Widgets
    add_btn: Gtk.Button = Gtk.Template.Child()
    scan_btn: Gtk.Button = Gtk.Template.Child()
    back_btn: Gtk.Button = Gtk.Template.Child()

    column: Handy.Column = Gtk.Template.Child()

    def __init__(self):
        super(AddAccountWindow, self).__init__()
        self.init_template('AddAccountWindow')
        self.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK)
        self.__init_widgets()

    def __init_widgets(self):
        """Create the Add Account widgets."""
        self.account_config = AccountConfig()
        self.account_config.connect("changed", self._on_account_config_changed)

        self.scan_btn.connect("clicked", self.account_config.scan_qr)
        self.back_btn.connect("clicked", lambda *_: self.destroy())
        self.column.add(self.account_config)

    def _on_account_config_changed(self, _, state: bool):
        """Set the sensitivity of the AddButton depends on the AccountConfig."""
        self.add_btn.set_sensitive(state)

    @Gtk.Template.Callback('add_btn_clicked')
    def _on_add(self, *_):
        account_obj = self.account_config.account
        # Create a new account
        account = Account.create(account_obj["username"],
                                 account_obj["token"],
                                 account_obj["provider"].provider_id)
        # Add it to the AccountsManager
        AccountsManager.get_default().add(account_obj["provider"], account)
        AccountsWidget.get_default().append(account)
        self.destroy()


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/account_config.ui')
class AccountConfig(Gtk.Overlay):
    __gtype_name__ = 'AccountConfig'
    # Signals
    __gsignals__ = {
        'changed': (
            GObject.SignalFlags.RUN_LAST,
            None, (bool,)
        ),
    }
    # Properties
    is_edit = GObject.Property(type=bool, default=False)
    # Widgets
    main_container: Gtk.Box = Gtk.Template.Child()

    proivder_image: ProviderImage

    provider_combobox = Gtk.Template.Child()
    providers_store = Gtk.Template.Child()
    provider_entry: Gtk.Entry = Gtk.Template.Child()

    account_name_entry: Gtk.Entry = Gtk.Template.Child()
    provider_website_entry: Gtk.Entry = Gtk.Template.Child()
    token_entry: Gtk.Entry = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super(AccountConfig, self).__init__()
        self.init_template('AccountConfig')

        self.props.is_edit = kwargs.get("edit", False)
        self._account = kwargs.get("account", None)
        self._notification = Notification()
        self.__init_widgets()

    @property
    def account(self):
        """
            Return an instance of Account for the new account.
        """
        provider_name = self.provider_entry.get_text()
        provider = Provider.get_by_name(provider_name)

        # Create a new provider if we don't find one
        if not provider:
            provider_image = self.provider_image.image
            provider_website = self.provider_website_entry.get_text()
            provider = Provider.create(provider_name, provider_website, None,
                                       provider_image)
        # Update the provider image if it changed
        elif provider and self.provider_image.image != provider.image:
            provider.update(image=self.provider_image.image)

        account = {
            "username": self.account_name_entry.get_text(),
            "provider": provider
        }
        if not self.props.is_edit:
            # remove spaces
            token = self.token_entry.get_text()
            account["token"] = "".join(token.split())
        return account

    def __init_widgets(self):
        self.add_overlay(self._notification)
        if self._account is not None:
            self.provider_image = ProviderImage(self._account.provider,
                                                96)
            self.token_entry.props.secondary_icon_activatable = self._account.provider.doc_url is not None
        else:
            self.token_entry.props.secondary_icon_activatable = False
            self.provider_image = ProviderImage(None, 96)

        self.main_container.pack_start(self.provider_image, False, False, 0)
        self.main_container.reorder_child(self.provider_image, 0)
        self.provider_image.set_halign(Gtk.Align.CENTER)

        # Set up auto completion
        if self._account and self._account.provider:
            self.provider_entry.set_text(self._account.provider.name)

        if self._account and self._account.username:
            self.account_name_entry.set_text(self._account.username)

        if self.props.is_edit:
            self.token_entry.hide()
            self.token_entry.set_no_show_all(True)
        else:
            self.token_entry.connect("icon-press", self.__on_open_doc_url)

        self._fill_data()

    def __on_open_doc_url(self, *args):
        provider_name = self.provider_entry.get_text()
        provider = Provider.get_by_name(provider_name)
        if provider and provider.doc_url:
            Gio.app_info_launch_default_for_uri(provider.doc_url)
        else:
            self.token_entry.props.secondary_icon_activatable = False

    @Gtk.Template.Callback('provider_changed')
    def _on_provider_changed(self, combo):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            provider_id = model[tree_iter][0]
            provider = Provider.get_by_id(provider_id)
        else:
            provider_name = self.provider_entry.get_text()
            provider = Provider.get_by_name(provider_name)
        # if we find a provider already saved on the database
        if provider:
            self.token_entry.props.secondary_icon_activatable = provider.doc_url is not None
            self.provider_image.emit("provider-changed", provider)
            self.provider_website_entry.hide()
            self.provider_website_entry.set_no_show_all(True)
        else:
            self.provider_website_entry.show()
            self.provider_website_entry.set_no_show_all(False)
            self.provider_image.set_state(ProviderImageState.NOT_FOUND)

    def _fill_data(self):
        providers = Provider.all()
        for provider in providers:
            self.providers_store.append([provider.provider_id, provider.name])

    @Gtk.Template.Callback('account_edited')
    def _validate(self, *_):
        """Validate the username and the token."""
        provider = self.provider_entry.get_text()
        username = self.account_name_entry.get_text()
        token = "".join(self.token_entry.get_text().split())

        if not username:
            self.account_name_entry.get_style_context().add_class("error")
            valid_name = False
        else:
            self.account_name_entry.get_style_context().remove_class("error")
            valid_name = True

        if not provider:
            self.provider_combobox.get_style_context().add_class("error")
            valid_provider = False
        else:
            self.provider_combobox.get_style_context().remove_class("error")
            valid_provider = True

        if (not token or not OTP.is_valid(token)) and not self.props.is_edit:
            self.token_entry.get_style_context().add_class("error")
            valid_token = False
        else:
            self.token_entry.get_style_context().remove_class("error")
            valid_token = True

        self.emit("changed", all([valid_name, valid_provider, valid_token]))

    @Gtk.Template.Callback('on_provider_website_changed')
    def on_provider_website_changed(self, entry, event):
        '''Update the website favicon once the URL is updated'''
        if entry.get_visible():
            website = entry.get_text().strip()
            self.provider_image.fetch_favicon_from_url(website)

    def scan_qr(self, *args):
        '''Scans a QRCode and fills the entries with the correct data.'''
        try:
            filename = GNOMEScreenshot.area()
            assert filename
            account = QRReader.from_file(filename)
            assert account is dict
            self.token_entry.set_text(account.get('token',
                                                  self.token_entry.get_text()))
            self.provider_entry.set_text(account.get('provider',
                                                     self.provider_entry.get_text()))
            self.account_name_entry.set_text(account.get('username',
                                                         self.account_name_entry.get_text()))
        except AssertionError:
            self._notification.send(_("Invalid QR code"),
                                    timeout=3)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gettext import gettext as _
from gi.repository import Gtk, GObject, GLib
from .edit import EditAccountWindow


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/account_row.ui')
class AccountRow(Gtk.ListBoxRow):
    """
        AccountRow Widget.

        It's a subclass of Gtk.ListBoxRow
        Added as a child to a AccountsList

        @signals: None
        @properties: account
    """
    __gtype_name__ = 'AccountRow'

    __gsignals__ = {
        'pin-copied': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        ),
        'account-updated': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        ),
    }

    account_name_label: Gtk.Label = Gtk.Template.Child()
    pin_label: Gtk.Label = Gtk.Template.Child()

    more_actions_btn: Gtk.Button = Gtk.Template.Child()
    delete_btn: Gtk.Button = Gtk.Template.Child()
    copy_btn_stack: Gtk.Stack = Gtk.Template.Child()

    _timeout_id: int = 0

    def __init__(self, account):
        """
        :param account: Account
        """
        super(AccountRow, self).__init__()
        self.init_template('AccountRow')
        self._account = account

        self._account.connect("otp_updated", self._on_pin_updated)
        self.__init_widgets()

    @property
    def account(self):
        """
            The Account model assigned to this AccountRow

            :return: Account Object
        """
        return self._account

    def __init_widgets(self):
        # Set up account name text label
        self.account_name_label.set_text(self.account.username)
        self.account_name_label.set_tooltip_text(self.account.username)

        # Set up account pin text label
        pin = self.account.otp.pin
        if pin:
            self.pin_label.set_text(pin)
        else:
            self.pin_label.set_text("??????")
            self.pin_label.set_tooltip_text(_("Couldn't generate the secret code"))

    @Gtk.Template.Callback('copy_btn_clicked')
    def _on_copy(self, *__):
        """
            Copy button clicked signal handler.
            Copies the OTP pin to the clipboard
        """
        self.copy_btn_stack.set_visible_child_name("ok_image")
        self._account.copy_pin()
        self.emit("pin-copied", _("The PIN of {} was copied to the clipboard").format(self.account.username))

        def btn_clicked_timeout_callback(*_):
            self.copy_btn_stack.set_visible_child_name("copy_image")
            if self._timeout_id > 0:
                GLib.Source.remove(self._timeout_id)
                self._timeout_id = 0

        self._timeout_id = GLib.timeout_add_seconds(2,
                                                    btn_clicked_timeout_callback,
                                                    None)

    @Gtk.Template.Callback('edit_btn_clicked')
    def _on_edit(self, *_):
        """
            Edit Button clicked signal handler.
            Opens a new Window to edit the current account.
        """
        from Authenticator.widgets import Window
        main_window = Window.get_default()
        edit_window = EditAccountWindow(self._account)
        edit_window.set_transient_for(main_window)
        edit_window.set_size_request(*main_window.get_size())
        edit_window.resize(*main_window.get_size())
        edit_window.connect("updated", self._on_update)
        edit_window.show_all()
        edit_window.present()

    def _on_update(self, __, account_name: str, provider):
        """
            On account update signal handler.
            Updates the account name and provider

            :param account_name: the new account's name
            :type account_name: str

            :param provider: the new account's provider
            :type provider: str
        """
        self.account_name_label.set_text(account_name)
        self.account.update(account_name, provider)
        self.emit("account-updated", _("The account was updated successfully"))

    def _on_pin_updated(self, _, pin: str):
        """
            Updates the pin label each time a new OTP is generated.
            otp_updated signal handler.

            :param pin: the new OTP
            :type pin: str
        """
        if pin:
            self.pin_label.set_text(pin)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gettext import gettext as _
from gi.repository import GObject, Gtk, Gdk
from hashlib import sha256
from typing import Union
from Authenticator.models import Database, Keyring, Logger, OTP, Provider


class Account(GObject.GObject):

    __gsignals__ = {
        'otp_out_of_date': (
            GObject.SignalFlags.RUN_LAST,
            None,
            ()
        ),
        'otp_updated': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str, )
        ),
        'removed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            ()
        ),
    }
    _provider: Provider = None

    def __init__(self, _id: str, username: str, token_id: str, provider: int):
        GObject.GObject.__init__(self)
        self.id = _id
        self.username = username
        self.provider = provider
        self._token_id = token_id
        token = Keyring.get_default().get_by_id(self._token_id)
        self.connect("otp_out_of_date", self._on_otp_out_of_date)
        if token:
            self.otp = OTP(token)
            self._code_generated = True
        else:
            self.otp = None
            self._code_generated = False
            Logger.error("Could not read the secret code,"
                         "the keyring keys were reset manually")

    @staticmethod
    def create(username: str, token: str, provider: int) -> 'Account':
        """
        Create a new Account.
        :param username: the account's username
        :param provider: the account's provider
        :param token: the OTP secret token
        :return: Account object
        """
        # Encrypt the token to create a secret_id
        token_id = sha256(token.encode('utf-8')).hexdigest()
        # Save the account
        obj = Database.get_default().insert_account(username, token_id, provider)
        Keyring.get_default().insert(token_id, provider, username, token)
        return Account(obj.id, username, token_id, provider)

    @staticmethod
    def create_from_json(json_obj: dict) -> 'Account':
        tags = json_obj["tags"]
        if not tags:
            provider_name = _("Default")
        else:
            provider_name = tags[0]
        provider = Provider.get_by_name(provider_name)
        if not provider:
            provider = Provider.create(provider_name, None, None, None)
        return Account.create(json_obj["label"], json_obj["secret"], provider.provider_id)

    @staticmethod
    def get_by_id(id_: int) -> 'Account':
        obj = Database.get_default().account_by_id(id_)
        return Account(obj.id, obj.username, obj.token_id, obj.provider)

    @property
    def provider(self) -> 'Provider':
        return self._provider

    @provider.setter
    def provider(self, provider: Union[int, 'Provider']) -> 'Provider':
        if isinstance(provider, int):
            self._provider = Provider.get_by_id(provider)
        else:
            self._provider = provider

    def update(self, username: str, provider: Provider):
        """
        Update the account name and/or provider.
        :param username: the account's username
        :param provider: the account's provider
        """
        self.username = username
        self.provider = provider
        account = {
            "username": username,
            "provider": provider.provider_id,
        }
        Database.get_default().update_account(account, self.id)

    def remove(self):
        """
        Remove the account.
        """
        Database.get_default().delete_account(self.id)
        Keyring.get_default().remove(self._token_id)
        self.emit("removed")
        Logger.debug("Account '{}' with id {} was removed".format(self.username,
                                                                  self.id))

    def copy_pin(self):
        """Copy the OTP to the clipboard."""

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(self.otp.pin, -1)

    def _on_otp_out_of_date(self, *_):
        if self._code_generated:
            self.otp.update()
            self.emit("otp_updated", self.otp.pin)

    def to_json(self):
        token = Keyring.get_default().get_by_id(self._token_id)
        if token:
            return {
                "secret": token,
                "label": self.username,
                "period": 30,
                "digits": 6,
                "type": "OTP",
                "algorithm": "SHA1",
                "thumbnail": "Default",
                "last_used": 0,
                "tags": [self.provider.name]
            }
        return {}



#!/usr/bin/env python3
"""
YAML database to JSON converter.
"""
import json
import tempfile
from collections import OrderedDict
from glob import glob
from os import path, remove
from shutil import rmtree
from subprocess import call
import sys
try:
    import yaml
except ImportError:
    sys.exit("Please install pyaml first")

try:
    from HTMLParser import HTMLParser
except ImportError:
    from html.parser import HTMLParser



GIT_CLONE_URI = "https://github.com/2factorauth/twofactorauth"
TMP_FOLDER = path.join(tempfile.gettempdir(), "Authenticator")
DATA_DIR = path.join(TMP_FOLDER, "_data")
OUTPUT_DIR = path.join(path.dirname(
    path.realpath(__file__)), "../data/data.json")

print("Cloning the repository...")
if path.exists(TMP_FOLDER):
    rmtree(TMP_FOLDER)
call(["git", "clone", "--depth=1", GIT_CLONE_URI, TMP_FOLDER])

if path.exists(OUTPUT_DIR):
    remove(OUTPUT_DIR)


def is_valid(provider):
    if set(["tfa", "software"]).issubset(provider.keys()):
        return provider["tfa"] and provider["software"]
    else:
        return False


output = {}

html_parser = HTMLParser()

for db_file in glob(DATA_DIR + "/*.yml"):
    with open(db_file, 'r', encoding='utf8') as file_data:
        try:
            providers = yaml.load(file_data)["websites"]
            for provider in providers:
                if is_valid(provider):
                    name = provider.get("name")
                    output[html_parser.unescape(name)] = {
                        "img": provider.get("img"),
                        "url": provider.get("url", ""),
                        "doc": provider.get("doc", "")
                    }
        except (yaml.YAMLError, TypeError, KeyError) as error:
            pass
output = OrderedDict(sorted(output.items(), key=lambda x: x[0].lower()))

with open(OUTPUT_DIR, 'w') as data:
    json.dump(output, data, indent=4)

rmtree(TMP_FOLDER)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gi.repository import GObject, Secret


class Keyring(GObject.GObject):
    ID: str = "com.github.bilelmoussaoui.Authenticator"
    PasswordID: str = "com.github.bilelmoussaoui.Authenticator.Login"
    PasswordState: str = "com.github.bilelmoussaoui.Authenticator.State"
    instance: 'Keyring' = None

    can_be_locked: GObject.Property = GObject.Property(type=bool, default=False)

    def __init__(self):
        GObject.GObject.__init__(self)
        self.schema = Secret.Schema.new(Keyring.ID,
                                        Secret.SchemaFlags.NONE,
                                        {
                                            "id": Secret.SchemaAttributeType.STRING,
                                            "name": Secret.SchemaAttributeType.STRING,
                                        })
        self.password_schema = Secret.Schema.new(Keyring.PasswordID,
                                                 Secret.SchemaFlags.NONE,
                                                 {"password": Secret.SchemaAttributeType.STRING})
        self.password_state_schema = Secret.Schema.new(Keyring.PasswordState,
                                                       Secret.SchemaFlags.NONE,
                                                       {"state": Secret.SchemaAttributeType.STRING})
        self.props.can_be_locked = self.is_password_enabled() and self.has_password()

    @staticmethod
    def get_default():
        if Keyring.instance is None:
            Keyring.instance = Keyring()
        return Keyring.instance

    def get_by_id(self, token_id: str) -> str:
        """
        Return the OTP token based on a secret ID.

        :param token_id: the secret ID associated to an OTP token
        :type token_id: str
        :return: the secret OTP token.
        """
        schema = self.schema
        token = Secret.password_lookup_sync(schema, {"id": str(token_id)},
                                            None)
        return token

    def insert(self, token_id: str, provider: str, username: str, token: str):
        """
        Save a secret OTP token.

        :param token_id: The secret ID associated to the OTP token
        :param provider: the provider name
        :param username: the username
        :param token: the secret OTP token.


        """
        schema = self.schema

        data = {
            "id": str(token_id),
            "name": str(username),
        }
        Secret.password_store_sync(
            schema,
            data,
            Secret.COLLECTION_DEFAULT,
            "{provider} OTP ({username})".format(provider=provider,
                                                 username=username),
            token,
            None
        )

    def remove(self, token_id: str) -> bool:
        """
        Remove a specific secret OTP token.

        :param secret_id: the secret ID associated to the OTP token
        :return bool: Either the token was removed successfully or not
        """
        schema = self.schema
        success = Secret.password_clear_sync(schema, {"id": str(token_id)},
                                             None)
        return success

    def clear(self) -> bool:
        """
           Clear all existing accounts.

           :return bool: Either the token was removed successfully or not
       """
        schema = self.schema
        success = Secret.password_clear_sync(schema, {}, None)
        return success

    def get_password(self) -> str:
        schema = self.password_schema
        password = Secret.password_lookup_sync(schema, {}, None)
        return password

    def set_password(self, password: str):
        schema = self.password_schema
        # Clear old password
        self.remove_password()
        # Store the new one
        Secret.password_store_sync(
            schema,
            {},
            Secret.COLLECTION_DEFAULT,
            "Authenticator password",
            password,
            None
        )
        self.set_password_state(True)

    def is_password_enabled(self) -> bool:
        schema = self.password_state_schema
        state = Secret.password_lookup_sync(schema, {}, None)
        return state == 'true' if state else False

    def set_password_state(self, state: bool):
        schema = self.password_state_schema
        if not state:
            Secret.password_clear_sync(schema, {}, None)
        else:
            Secret.password_store_sync(
                schema,
                {},
                Secret.COLLECTION_DEFAULT,
                "Authenticator state",
                "true",
                None
            )
        self.props.can_be_locked = state and self.has_password()

    def has_password(self) -> bool:
        return self.get_password() is not None

    def remove_password(self):
        schema = self.password_schema
        Secret.password_clear_sync(schema, {}, None)
        self.set_password_state(False)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You  ould have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
import sqlite3
from os import path, makedirs
from gi.repository import GLib
from collections import namedtuple
from shutil import move
from typing import Iterable

from Authenticator.models import Logger


Provider = namedtuple('Provider', ['id', 'name', 'website', 'doc_url', 'image'])
Account = namedtuple('Account', ['id', 'username', 'token_id', 'provider'])


class Database:
    """SQL database handler."""

    # Default instance
    instance = None
    # Database version number
    db_version: int = 7

    def __init__(self):
        self.migrations_dir = path.join(path.dirname(__file__), '../migrations')
        database_created = self.__create_database_file()
        if database_created:
            self.__apply_migrations()
        self.conn = sqlite3.connect(self.db_file)

    @staticmethod
    def get_default():
        """Return the default instance of database"""
        if Database.instance is None:
            Database.instance = Database()
        return Database.instance

    @property
    def db_dir(self) -> str:
        return path.join(GLib.get_user_config_dir(),
                         'Authenticator')

    @property
    def db_file(self) -> str:
        return path.join(self.db_dir,
                         'database-{}.db'.format(str(Database.db_version))
                         )

    def insert_account(self, username: str, token_id: str, provider: str):
        """
        Insert a new account to the database
        :param username: Account name
        :param token_id: The token identifier stored using libsecret
        :param provider: The provider foreign key
        """
        query = "INSERT INTO accounts (username, token_id, provider) VALUES (?, ?, ?)"
        cursor = self.conn.cursor()
        try:
            cursor.execute(query, [username, token_id, provider])
            self.conn.commit()
            return Account(cursor.lastrowid, username, token_id, provider)
        except Exception as error:
            Logger.error("[SQL] Couldn't add a new account")
            Logger.error(str(error))

    def insert_provider(self, name: str, website: str,
                        doc_url: str = None, image: str = None):
        """
        Insert a new provider to the database
        :param name: The provider name
        :param website: The provider website, used to extract favicon
        :param doc_url: The provider doc_url, used to allow user get the docs
        :param image: The image path of a provider
        """
        query = "INSERT INTO providers (name, website, doc_url, image) VALUES (?, ?, ?, ?)"
        cursor = self.conn.cursor()
        try:
            cursor.execute(query, [name, website, doc_url, image])
            self.conn.commit()
            return Provider(cursor.lastrowid, name, website, doc_url, image)
        except Exception as error:
            Logger.error("[SQL] Couldn't add a new account")
            Logger.error(str(error))

    def account_by_id(self, id_: int) -> Account:
        """
            Get an account by the ID
            :param id_: int the account id
            :return: Account: The account data
        """
        query = "SELECT * FROM accounts WHERE id=?"
        try:
            data = self.conn.cursor().execute(query, (id_,))
            return Account(*data.fetchone())
        except Exception as error:
            Logger.error("[SQL] Couldn't get account with ID={}".format(id_))
            Logger.error(str(error))
        return None

    def accounts_by_provider(self, provider_id: int) -> Iterable[Account]:
        query = "SElECT * FROM accounts WHERE provider=?"
        query_d = self.conn.execute(query, (provider_id, ))
        accounts = query_d.fetchall()
        return [Account(*account) for account in accounts]

    def provider_by_id(self, id_: int) -> Provider:
        """
            Get a provider by the ID
            :param id_: int the provider id
            :return: Provider: The provider data
        """
        query = "SELECT * FROM providers WHERE id=?"
        try:
            data = self.conn.cursor().execute(query, (id_,))
            return Provider(*data.fetchone())
        except Exception as error:
            Logger.error("[SQL] Couldn't get provider with ID={}".format(id_))
            Logger.error(str(error))
        return None

    def provider_by_name(self, provider_name: str) -> Provider:
        """
            Get a provider by the ID
            :param id_: int the provider id
            :return: Provider: The provider data
        """
        query = "SELECT * FROM providers WHERE name LIKE ? "
        try:
            data = self.conn.cursor().execute(query, (provider_name,))
            provider = data.fetchone()
            return Provider(*provider) if provider else None
        except Exception as error:
            Logger.error("[SQL] Couldn't get provider with name={}".format(provider_name))
            Logger.error(str(error))
        return None

    def delete_account(self, id_: int):
        """
            Remove an account by ID.

            :param id_: the account ID
            :type id_: int
        """
        self.__delete("accounts", id_)

    def delete_provider(self, id_: int):
        """
            Remove a provider by ID.

            :param id_: the provider ID
            :type id_: int
        """
        self.__delete("providers", id_)

    def update_account(self, account_data: dict, id_: int):
        """
        Update an account by id
        """
        self.__update_by_id("accounts", account_data, id_)

    def update_provider(self, provider_data: dict, id_: int):
        # Update a provider by id
        self.__update_by_id("providers", provider_data, id_)

    def search_accounts(self, terms: Iterable[str]) -> Iterable[Account]:
        if terms:
            filters = " ".join(terms)
            if filters:
                filters = "%" + filters + "%"
            query = """
                        SELECT A.* FROM accounts A
                        JOIN providers P
                        ON A.provider = P.id
                        WHERE
                        A.username LIKE ?
                        OR
                        P.name LIKE ?
                        GROUP BY provider
                        ORDER BY  A.username ASC
                    """
            try:
                data = self.conn.cursor().execute(query, (filters, filters, ))
                accounts = data.fetchall()
                return [Account(*account) for account in accounts]
            except Exception as error:
                Logger.error("[SQL]: Couldn't search for an account")
                Logger.error(str(error))
        return []

    @property
    def accounts_count(self):
        """
            Count the total number of existing accounts.

           :return: int
        """
        return self.__count("accounts")

    @property
    def providers_count(self):
        """
            Count the total number of existing providers
            :return: int
        """
        return self.__count("providers")

    def get_providers(self, **kwargs) -> Iterable[Provider]:
        only_used = kwargs.get("only_used",)
        query = "SELECT * FROM providers"
        if only_used:
            query += " WHERE id IN (SELECT DISTINCT provider FROM accounts)"
        try:
            data = self.conn.cursor().execute(query)
            providers = data.fetchall()
            return [Provider(*provider) for provider in providers]
        except Exception as error:
            Logger.error("[SQL] Couldn't fetch providers list")
            Logger.error(str(error))
        return None

    @property
    def accounts(self) -> [Account]:
        """
            Retrieve the list of accounts.

            :return list
        """
        query = "SELECT * FROM accounts"
        try:
            data = self.conn.cursor().execute(query)
            accounts = data.fetchall()
            return [Account(*account) for account in accounts]
        except Exception as error:
            Logger.error("[SQL] Couldn't fetch accounts list")
            Logger.error(str(error))
        return None

    def __create_database_file(self):
        """
        Create an empty database file for the first start of the application.
        """
        created = False
        for i in range(3, self.db_version + 1):
            db_file = path.join(self.db_dir, "database-{}.db".format(i))
            if path.exists(db_file):
                move(db_file, self.db_file)
                created = True
                break
        if not created:
            makedirs(path.dirname(self.db_file), exist_ok=True)
            if not path.exists(self.db_file):
                with open(self.db_file, 'w') as file_obj:
                    file_obj.write('')
                created = True
        return created

    def __apply_migrations(self):
        """
        Create the needed tables to run the application.
        """
        from yoyo import read_migrations
        from yoyo import get_backend
        backend = get_backend('sqlite:///' + self.db_file)
        migrations = read_migrations(self.migrations_dir)
        with backend.lock():
            backend.apply_migrations(backend.to_apply(migrations))

    def __count(self, table_name: str) -> int:
        query = "SELECT COUNT(id) AS count FROM " + table_name
        try:
            data = self.conn.cursor().execute(query)
            return data.fetchone()[0]
        except Exception as error:
            Logger.error("[SQL]: Couldn't count the results from " + table_name)
            Logger.error(str(error))
        return None

    def __delete(self, table_name: str, id_: int):
        """
            Remove a row by ID.

            :param id_: the row ID
            :type id_: int
        """
        query = "DELETE FROM {} WHERE id=?".format(table_name)
        try:
            self.conn.execute(query, (id_,))
            self.conn.commit()
        except Exception as error:
            Logger.error("[SQL] Couldn't remove the row '{}'".format(id_))
            Logger.error(str(error))

    def __update_by_id(self, table_name: str, data: dict, id_: int):
        query = "UPDATE {} SET ".format(table_name)
        resources = []
        i = 0
        for table_column, value in data.items():
            query += " {}=?".format(table_column)
            if i != len(data) - 1:
                query += ","
            resources.append(value)
            i += 1
        resources.append(id_)
        query += "WHERE id=?"
        try:
            self.conn.execute(query, resources)
            self.conn.commit()
        except Exception as error:
            Logger.error("[SQL] Couldn't update row by id")
            Logger.error(error)



"""
Restore old accounts
"""

from yoyo import step
from collections import OrderedDict

__depends__ = {'authenticator_20190525_03_R7miN-add-default-providers'}


def do_step(conn):
    # atler columns from secret id to token_id before
    # to fix the issue on older db
    try:
        accounts = conn.execute("SELECT id, username, token_id, provider FROM accounts").fetchall()
    except Exception:
        accounts = conn.execute("SELECT id, username, secret_id, provider FROM accounts").fetchall()
    _accounts = []

    providers_db = conn.execute("SELECT id, name FROM providers").fetchall()
    providers = OrderedDict()
    for provider_id, provider_name in providers_db:
        providers[provider_id] = provider_name.lower()
    cursor = conn.cursor()

    for account_id, username, secret_id, provider in accounts:
        if isinstance(provider, str):
            if provider.lower() not in providers.values():
                cursor.execute("INSERT INTO providers (name) VALUES (?)", (provider, ))
                provider_id = cursor.lastrowid
            else:
                provider_index = list(providers.values()).index(provider.lower())
                provider_id = list(providers.keys())[provider_index]
        else:
            provider_id = provider
        _accounts.append((account_id, username, provider_id, secret_id))

    cursor.execute(" ALTER TABLE accounts RENAME TO tmp;")
    cursor.execute(''' CREATE TABLE "accounts" (
                    "id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL UNIQUE,
                    "username" VARCHAR NOT NULL,
                    "token_id" VARCHAR NOT NULL UNIQUE,
                    "provider" INTEGER NOT NULL
                    )''')
    cursor.execute("DROP TABLE tmp;")
    added_tokens = []
    for account_id, username, provider_id, token_id in _accounts:
        if token_id not in added_tokens:
            cursor.execute("INSERT INTO accounts (username, provider, token_id) VALUES (?, ?, ?)", (username, provider_id, token_id))
            added_tokens.append(token_id)


steps = [
    step(do_step, ignore_errors='apply')
]



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
import binascii
from pyotp import TOTP


class OTP(TOTP):
    """
        OTP (One-time password) handler using PyOTP.
    """

    def __init__(self, token: str):
        """
        :param token: the OTP token.
        """
        TOTP.__init__(self, token)
        self.pin = None
        self.update()

    @staticmethod
    def is_valid(token: str) -> bool:
        """
        Validate a OTP token.

        :param token: OTP token
        :type token: str

        :return: bool
        """
        try:
            TOTP(token).now()
            return True
        except (binascii.Error, ValueError, TypeError):
            return False

    def update(self):
        """
            Generate a new OTP based on the same token.
        """
        try:
            self.pin = self.now()
        except binascii.Error:
            self.pin = None



"""
Add default providers
"""
from gi.repository import Gio
from yoyo import step
import json
__depends__ = {'authenticator_20190525_02_mdR2o-create-table-providers'}


def do_step(conn):
    uri = 'resource:///com/github/bilelmoussaoui/Authenticator/data.json'
    g_file = Gio.File.new_for_uri(uri)
    content = str(g_file.load_contents(None)[1].decode("utf-8"))
    data = json.loads(content)
    providers = []

    providers_db = conn.execute("SELECT name FROM providers").fetchall()
    providers_db = [provider[0].lower() for provider in providers_db]

    for provider_name, provider_info in data.items():
        if not provider_name.lower() in providers_db:
            providers.append((provider_name, provider_info['url'],
                              provider_info['doc'],))
    query = "INSERT INTO providers (name, website, doc_url) VALUES (?, ?, ?)"
    conn.executemany(query, providers)
    conn.commit()


steps = [
    step(do_step, ignore_errors='apply')
]



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from os import path
from gi.repository import Gtk

from Authenticator.models import Database


class Provider:

    instance: 'Provider' = None

    def __init__(self,
                 provider_id: int = None,
                 name: str = None,
                 website: str = None,
                 doc_url: str = None,
                 image: str = None):

        self.provider_id = provider_id
        self.name = name
        self.website = website
        self.doc_url = doc_url
        self.image = image

    @staticmethod
    def create(name: str, website: str, doc_url: str, image: str) -> 'Provider':
        provider = Database.get_default().insert_provider(name, website, doc_url, image)
        return Provider(*provider)

    @staticmethod
    def get_by_id(id_) -> 'Provider':
        provider = Database.get_default().provider_by_id(id_)
        return Provider(*provider) if provider else None

    @staticmethod
    def get_by_name(name) -> 'Provider':
        provider = Database.get_default().provider_by_name(name)
        return Provider(*provider) if provider else None

    @staticmethod
    def all() -> ['Provider']:
        providers = Database.get_default().get_providers()
        return [
            Provider(*provider)
            for provider in providers
        ]

    @property
    def image_path(self) -> str:
        from Authenticator.widgets import ProviderImage
        cached_icon = path.join(ProviderImage.CACHE_DIR, self.image)
        if path.exists(cached_icon):
            return cached_icon
        else:
            theme = Gtk.IconTheme.get_default()
            icon_info = theme.lookup_icon("image-missing", 48, 0)
            if icon_info:
                return icon_info.get_filename()
        return None

    def update(self, **provider_data):
        self.name = provider_data.get("name", self.name)
        self.image = provider_data.get("image", self.image)
        Database.get_default().update_provider(provider_data, self.provider_id)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""

from gi.repository import Gtk, GObject, GdkPixbuf, GLib, Gio
from os import path
from pathlib import Path
from enum import Enum
from tempfile import NamedTemporaryFile
import asyncio
from pyfavicon import Favicon
import concurrent
from Authenticator.models import Provider


class ProviderImageState(Enum):
    FOUND = 0
    NOT_FOUND = 1
    LOADING = 2

    @property
    def stack_name(self):
        if self.value == 0:
            return "provider_image"
        elif self.value == 1:
            return "provider_not_found"
        else:
            return "provider_loading"


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/provider_image.ui')
class ProviderImage(Gtk.Stack):
    """
        An Object that represents the provider service's image

        It contains a Gtk.Spinner and a Gtk.Image inside a Gtk.Stack
        It shows the Spinner when the provider image is being downloaded,
        once the download was successful, we show that image
    """
    __gtype_name__ = 'ProviderImage'

    __gsignals__ = {
        'provider-changed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT, )
        ),
        'image-downloaded': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str, GObject.TYPE_PYOBJECT, )
        )
    }

    provider_image = Gtk.Template.Child()
    provider_spinner = Gtk.Template.Child()
    image_eventbox = Gtk.Template.Child()
    insert_image = Gtk.Template.Child()
    not_found_box = Gtk.Template.Child()

    _timeout_id = 0
    CACHE_DIR = path.join(GLib.get_user_cache_dir(), "Authenticator")

    def __init__(self, provider=None, image_size=48):
        super(ProviderImage, self).__init__()
        self.init_template('ProviderImage')
        self.provider = provider if provider else Provider()

        self.image_size = image_size

        self._build_widget()

    def _build_widget(self):
        self.set_state(ProviderImageState.NOT_FOUND)
        self.not_found_box.props.width_request = self.image_size
        self.not_found_box.props.height_request = self.image_size
        self.provider_image.set_pixel_size(self.image_size)

        self.connect("provider-changed", self.__on_provider_changed)

        def set_show_insert_image(state):
            if state != self.insert_image.get_visible():
                self.insert_image.set_visible(state)
                self.insert_image.set_no_show_all(not state)

        self.image_eventbox.connect("enter-notify-event",
                                    lambda *_: set_show_insert_image(True))
        self.image_eventbox.connect("leave-notify-event",
                                    lambda *_: set_show_insert_image(False))

        if self.provider.image:
            if not self.set_image(self.provider.image) and self.provider.website:
                self.fetch_favicon_from_url(self.provider.website)

    def set_state(self, state: ProviderImageState):
        self.state = state
        if state == ProviderImageState.LOADING:
            self.provider_spinner.start()
        else:
            self.provider_spinner.stop()
        self.set_visible_child_name(state.stack_name)

    @property
    def image(self):
        if self.provider:
            return self.provider.image
        return None

    def set_image(self, image):
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(image)
            pixbuf = pixbuf.scale_simple(self.image_size, self.image_size,
                                         GdkPixbuf.InterpType.BILINEAR)
            self.provider_image.set_from_pixbuf(pixbuf)
            self.set_state(ProviderImageState.FOUND)
            return True
        except GLib.Error:
            pass
        self.set_state(ProviderImageState.NOT_FOUND)
        return False

    # Callbacks
    @Gtk.Template.Callback('select_image_clicked')
    def __on_select_image_clicked(self, *args):
        dialog = Gtk.FileChooserNative()
        dialog.set_action(Gtk.FileChooserAction.OPEN)
        dialog.set_transient_for(self.get_toplevel())

        # Allow to select images only

        mimes = ('image/jpeg', 'image/png', 'image/svg+xml')
        for mime_type in mimes:
            img_filter = Gtk.FileFilter()
            img_filter.set_name(mime_type)
            img_filter.add_mime_type(mime_type)
            dialog.add_filter(img_filter)

        if dialog.run() == Gtk.ResponseType.ACCEPT:
            file_uri = dialog.get_uri()
            cache_file = self.__create_cache_file(file_uri)
            self.set_image(cache_file)
        dialog.destroy()

    # Signals

    def __on_provider_changed(self, _, provider):
        self.provider = provider
        self.provider_spinner.start()
        self.set_state(ProviderImageState.LOADING)
        if provider.image:  # If we have already an image in the database
            if path.exists(provider.image):
                self.set_image(provider.image)
            elif provider.website:
                self.fetch_favicon_from_url(provider.website)
            else:
                self.provider.update(image=None)
                self.set_state(ProviderImageState.NOT_FOUND)
        elif provider.website:  # If we can download a favicon
            self.fetch_favicon_from_url(provider.website)
        else:
            self.set_state(ProviderImageState.NOT_FOUND)

    def fetch_favicon_from_url(self, provider_website):
        if provider_website:
            worker_loop = asyncio.new_event_loop()
            pool = concurrent.futures.ThreadPoolExecutor()
            current_provider = self.provider
            pool.submit(self._download_favicon, worker_loop, provider_website,
                        current_provider)

    # Private
    def _download_favicon(self, loop, provider_website, current_provider):
        cache_dir = Path(ProviderImage.CACHE_DIR)
        if not cache_dir.exists():
            cache_dir.mkdir()

        async def fetch_favicon():
            favicon = Favicon(download_dir=cache_dir)
            icons = await favicon.from_url(provider_website)
            largest_icon = icons.get_largest()
            if largest_icon:
                await largest_icon.save()
                GLib.idle_add(self.emit, "image-downloaded", largest_icon.path,
                              current_provider)
            else:
                self.set_state(ProviderImageState.NOT_FOUND)
        # Hackish solution for https://github.com/python/cpython/pull/13548
        loop.set_exception_handler(lambda loop, ctx: None)
        loop.run_until_complete(fetch_favicon())

    def __create_cache_file(self, file_uri):
        """
            Store a copy of the image under the cache dir of Authenticator.
        """
        gfile = Gio.File.new_for_uri(file_uri)
        # Copy file to cache dir
        cache_filename = path.basename(NamedTemporaryFile().name)
        destination_file = path.join(ProviderImage.CACHE_DIR, cache_filename)
        dfile = Gio.File.new_for_path(destination_file)
        gfile.copy(dfile, Gio.FileCopyFlags.NONE, None, None, None)

        self.set_image(destination_file)
        return destination_file

    def do_image_downloaded(self, image_path, provider):

        if image_path and path.exists(image_path):
            if provider:
                provider.update(image=image_path)  # Update the database
            # If the user didn't change the provider while downloading
            if provider.provider_id == self.provider.provider_id:
                self.set_image(image_path)
                self.set_state(ProviderImageState.FOUND)
            return
        self.set_state(ProviderImageState.NOT_FOUND)



"""
Create table accounts
"""

from yoyo import step

__depends__ = {}

steps = [
    step('''
         CREATE TABLE "accounts" (
            "id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL UNIQUE,
            "username" VARCHAR NOT NULL,
            "provider" VARCHAR NOT NULL,
            "secret_id" VARCHAR NOT NULL UNIQUE)''',
         'DROP TABLE "accounts"',
         ignore_errors='apply')
]



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gettext import gettext as _
from gi.repository import Gtk

__all__ = ['import_json', 'export_json']


def import_json(parent: Gtk.Window) -> str:
    mimetype = {'type': "application/json", 'name': _("JSON files")}
    return __open_file_chooser(parent, mimetype)


def export_json(parent: Gtk.Window) -> str:
    mimetype = {'type': "application/json", 'name': _("JSON files")}
    return __open_file_chooser(parent, mimetype, Gtk.FileChooserAction.SAVE)


def __open_file_chooser(parent: Gtk.Window, mimetype: dict,
                        action=Gtk.FileChooserAction.OPEN) -> str:
    file_chooser = Gtk.FileChooserNative()
    file_chooser.set_action(action)
    file_chooser.set_transient_for(parent)
    filter_json = Gtk.FileFilter()
    filter_json.set_name(mimetype["name"])
    filter_json.add_mime_type(mimetype["type"])
    file_chooser.add_filter(filter_json)
    response = file_chooser.run()
    uri = None
    if response == Gtk.ResponseType.ACCEPT:
        uri = file_chooser.get_uri()
    file_chooser.destroy()
    return uri



"""
Create table providers
"""

from yoyo import step

__depends__ = {'authenticator_20190525_01_GdUDU-create-table-accounts'}

steps = [
    step(
        '''CREATE TABLE "providers" (
            "id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL UNIQUE,
            "name" VARCHAR NOT NULL,
            "website" VARCHAR NULL,
            "doc_url" VARCHAR NULL,
            "image" VARCHAR NULL
        )
        ''',
        'DROP TABLE "providers"',
        ignore_errors='apply'
    )
]



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
from gi.repository import Gtk, GObject, Gio, Handy

from Authenticator.models import Logger, Settings, AccountsManager
from Authenticator.widgets.accounts.add import AddAccountWindow
from Authenticator.widgets.accounts.list import AccountsWidget


class WindowView:
    NORMAL = 0
    LOCKED = 1
    EMPTY = 2


@Gtk.Template(resource_path='/com/github/bilelmoussaoui/Authenticator/window.ui')
class Window(Gtk.ApplicationWindow):
    """Main Window object."""
    __gtype_name__ = 'Window'

    # Default Window instance
    instance: 'Window' = None

    view = GObject.Property(type=int, default=0)

    search_btn: Gtk.ToggleButton = Gtk.Template.Child()
    primary_menu_btn: Gtk.MenuButton = Gtk.Template.Child()

    main_stack: Gtk.Stack = Gtk.Template.Child()
    headerbar_stack: Gtk.Stack = Gtk.Template.Child()
    accounts_stack: Gtk.Stack = Gtk.Template.Child()

    search_bar: Handy.SearchBar = Gtk.Template.Child()
    search_entry: Gtk.SearchEntry = Gtk.Template.Child()
    password_entry: Gtk.Entry = Gtk.Template.Child()

    def __init__(self):
        super(Window, self).__init__()
        self.init_template('Window')

        self.connect("notify::view", self.__state_changed)

        self.key_press_signal = None
        self.restore_state()
        # Start the Account Manager
        AccountsManager.get_default()

        self.__init_widgets()

    @staticmethod
    def get_default() -> 'Window':
        """Return the default instance of Window."""
        if Window.instance is None:
            Window.instance = Window()
        return Window.instance

    def close(self):
        self.save_state()
        AccountsManager.get_default().kill()
        self.destroy()

    def add_account(self, *_):
        if not self.get_application().is_locked:
            add_window = AddAccountWindow()
            add_window.set_transient_for(self)
            add_window.set_size_request(*self.get_size())
            add_window.resize(*self.get_size())
            add_window.show_all()
            add_window.present()

    def set_menu(self, menu: Gio.Menu):
        popover = Gtk.Popover.new_from_model(self.primary_menu_btn, menu)

        def primary_menu_btn_handler(_, popover):
            popover.set_visible(not popover.get_visible())

        self.primary_menu_btn.connect('clicked', primary_menu_btn_handler, popover)

    def toggle_search(self, *_):
        """
            Switch the state of the search mode

            Switches the state of the search mode if:
                - The application is not locked
                - There are at least one account in the database
            return: None
        """
        if self.props.view == WindowView.NORMAL:
            toggled = not self.search_btn.props.active
            self.search_btn.set_property("active", toggled)

    def refresh_view(self, *_):
        if AccountsManager.get_default().props.empty:
            self.props.view = WindowView.EMPTY
        else:
            self.props.view = WindowView.NORMAL

    def save_state(self):
        """
            Save window position and maximized state.
        """
        settings = Settings.get_default()
        settings.window_position = self.get_position()
        settings.window_maximized = self.is_maximized()

    def restore_state(self):
        """
            Restore the window's state.
        """
        settings = Settings.get_default()
        # Restore the window position
        position_x, position_y = settings.window_position
        if position_x != 0 and position_y != 0:
            self.move(position_x, position_y)
            Logger.debug("[Window] Restore position x: {}, y: {}".format(position_x,
                                                                         position_y))
        else:
            # Fallback to the center
            self.set_position(Gtk.WindowPosition.CENTER)

        if settings.window_maximized:
            self.maximize()

    def __init_widgets(self):
        """Build main window widgets."""
        # Register Actions
        self.__add_action("add-account", self.add_account)
        self.__add_action("toggle-searchbar", self.toggle_search)

        # Set up accounts Widget
        accounts_widget = AccountsWidget.get_default()
        accounts_widget.connect("account-removed", self.refresh_view)
        accounts_widget.connect("account-added", self.refresh_view)
        self.accounts_stack.add_named(accounts_widget, "accounts")
        self.accounts_stack.set_visible_child_name("accounts")

        AccountsManager.get_default().connect("notify::empty", self.refresh_view)

        self.search_bar.connect_entry(self.search_entry)
        self.search_bar.bind_property("search-mode-enabled", self.search_btn,
                                      "active",
                                      GObject.BindingFlags.BIDIRECTIONAL)

    def __add_action(self, key: str, callback, prop_bind: str = None,
                     bind_flag=GObject.BindingFlags.INVERT_BOOLEAN):
        action = Gio.SimpleAction.new(key, None)
        action.connect("activate", callback)
        if prop_bind:
            self.bind_property(prop_bind, action, "enabled", bind_flag)
        self.add_action(action)

    def __state_changed(self, *_):
        if self.props.view == WindowView.LOCKED:
            visible_child = "locked_state"
            visible_headerbar = "locked_headerbar"
            if self.key_press_signal:
                self.disconnect(self.key_press_signal)
        else:
            if self.props.view == WindowView.EMPTY:
                visible_child = "empty_state"
                visible_headerbar = "empty_headerbar"
            else:
                visible_child = "normal_state"
                visible_headerbar = "main_headerbar"
                # Connect on type search bar
                self.key_press_signal = self.connect("key-press-event",
                                                     lambda x, y: self.search_bar.handle_event(y))
        self.main_stack.set_visible_child_name(visible_child)
        self.headerbar_stack.set_visible_child_name(visible_headerbar)

    @Gtk.Template.Callback('unlock_btn_clicked')
    def __unlock_btn_clicked(self, *_):
        from Authenticator.models import Keyring
        typed_password = self.password_entry.get_text()
        if typed_password == Keyring.get_default().get_password():
            self.get_application().set_property("is-locked", False)
            # Reset password entry
            self.password_entry.get_style_context().remove_class("error")
            self.password_entry.set_text("")
        else:
            self.password_entry.get_style_context().add_class("error")

    @Gtk.Template.Callback('search_changed')
    def __search_changed(self, entry: Gtk.SearchEntry):
        """
            Handles search-changed signal.
        """
        def filter_func(row, data, *_):
            """
                Filter function
            """
            data = data.lower()
            if len(data) > 0:
                username = row.account.username.lower()
                provider_name = row.account.provider.name.lower()
                return (
                    data in username or data in provider_name
                )
            else:
                return True
        data = entry.get_text().strip()
        search_lists = AccountsWidget.get_default().accounts_lists
        results_count = 0
        for search_list in search_lists:
            search_list.set_filter_func(filter_func,
                                        data, False)
            for elem in search_list:

                if elem.get_child_visible():
                    results_count += 1

        if results_count == 0:
            self.accounts_stack.set_visible_child_name("empty_results")
        else:
            self.accounts_stack.set_visible_child_name("accounts")



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""

from gi.repository import Gtk, GObject, GLib
from gettext import gettext as _


class Notification(Gtk.Revealer):

    show_close_btn = GObject.Property(type=bool, default=False)
    show_action_btn = GObject.Property(type=bool, default=False)

    timeout = GObject.Property(type=int, default=5)
    action_callback = None

    _action_signal = None
    _source = 0

    def __init__(self):
        Gtk.Revealer.__init__(self)
        self._message = ""
        self._build_widget()
        self._bind_signals()

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, new_message):
        self._message = new_message
        self._notification_lbl.set_text(new_message)

    def send(self, message, **kwargs):
        self.message = message

        self._action_btn.set_label(kwargs.get("action_label", _("Undo")))

        self.props.show_action_btn = kwargs.get("show_action_btn", False)
        self.props.show_close_btn = kwargs.get("show_close_btn", False)
        self.props.timeout = kwargs.get("timeout", self.props.timeout)
        self.action_callback = kwargs.get("action_callback")

        self.set_reveal_child(True)
        self._source_id = GLib.timeout_add_seconds(self.timeout,
                                                   self._delete_notification, None)

    def _build_widget(self):
        self.set_transition_type(Gtk.RevealerTransitionType.SLIDE_DOWN)
        self.set_transition_duration(400)
        self.set_reveal_child(False)
        self.set_valign(Gtk.Align.START)
        self.set_halign(Gtk.Align.CENTER)

        frame = Gtk.Frame()
        frame.get_style_context().add_class("app-notification")
        notification_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self._close_btn = Gtk.Button()
        self._close_btn.connect("clicked", self._delete_notification)
        close_img = Gtk.Image.new_from_icon_name("window-close-symbolic", Gtk.IconSize.BUTTON)
        self._close_btn.get_style_context().add_class("flat")
        self._close_btn.set_tooltip_text(_("Close the notification"))
        self._close_btn.props.image = close_img
        self._close_btn.props_always_show_image = True

        self._action_btn = Gtk.Button()
        self._action_btn.set_label(_("Undo"))
        self._action_btn.connect("clicked", self._on_action_btn_clicked)

        self._notification_lbl = Gtk.Label()
        self._notification_lbl.set_text(self.message)

        notification_container.pack_start(self._notification_lbl, False, False, 6)
        notification_container.pack_end(self._close_btn, False, False, 6)
        notification_container.pack_end(self._action_btn, False, False, 6)

        frame.add(notification_container)
        self.add(frame)

    def _delete_notification(self, *args):
        self.set_reveal_child(False)
        self.message = ""
        if self._source_id > 0:
            GLib.Source.remove(self._source_id)

    def _on_action_btn_clicked(self, *args):
        if self.action_callback:
            self.action_callback()
        self._delete_notification()

    def _bind_signals(self):
        self._close_btn.bind_property("visible", self, "show-close-btn", GObject.BindingFlags.BIDIRECTIONAL)
        self._action_btn.bind_property("visible", self, "show-action-btn", GObject.BindingFlags.BIDIRECTIONAL)



"""
 Copyright © 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>

 This file is part of Authenticator.

 Authenticator is free software: you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published
 by the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Authenticator is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Authenticator. If not, see <http://www.gnu.org/licenses/>.
"""
import json
from gi.repository import Gio

from Authenticator.models import Account, AccountsManager, Logger
from Authenticator.widgets import AccountsWidget


class Backup:

    def __init__(self):
        pass

    @staticmethod
    def import_accounts(accounts: [dict]):
        accounts_widget = AccountsWidget.get_default()
        accounts_manager = AccountsManager.get_default()
        for account in accounts:
            try:
                new_account = Account.create_from_json(account)
                accounts_manager.add(new_account.provider, new_account)
                accounts_widget.append(new_account)
            except Exception as e:
                Logger.error("[Restore] Failed to import accounts")
                Logger.error(str(e))

    @staticmethod
    def export_accounts() -> [dict]:
        accounts = AccountsManager.get_default().accounts
        exported_accounts = []
        for account in accounts:
            json_account = account.to_json()
            exported_accounts.append(json_account)
        return exported_accounts


class BackupJSON:

    def __init__(self):
        pass

    @staticmethod
    def export_file(uri: str):
        accounts = Backup.export_accounts()
        gfile = Gio.File.new_for_uri(uri)
        stream = gfile.replace(None,
                               False,
                               Gio.FileCreateFlags.REPLACE_DESTINATION,
                               None)
        data_stream = Gio.DataOutputStream.new(stream)
        data_stream.put_string(json.dumps(accounts), None)
        stream.close()

    @staticmethod
    def import_file(uri: str):
        gfile = Gio.File.new_for_uri(uri)
        accounts = gfile.load_contents()[1].decode("utf-8")
        Backup.import_accounts(json.loads(accounts))
